# -*- coding: utf-8 -*-

from __future__ import print_function

import json
import csv
import codecs
import decimal
import datetime

from collections import namedtuple

try:
    import jinja2
except ImportError:
    from .common import MissingPackage
    jinja2 = MissingPackage("jinja2", "Templating engine")

from .errors import ArgumentError
from . import compat
from . import ext

from .browser import SPLIT_DIMENSION_NAME


__all__ = [
    "create_formatter",
    "CrossTableFormatter",
    "HTMLCrossTableFormatter",
    "SlicerJSONEncoder",
    "csv_generator",
    "JSONLinesGenerator",
]

def create_formatter(type_, *args, **kwargs):
    """Creates a formatter of type `type`. Passes rest of the arguments to the
    formatters initialization method."""
    return ext.formatter(type_, *args, **kwargs)


def _jinja_env():
    """Create and return cubes jinja2 environment"""
    loader = jinja2.PackageLoader('cubes', 'templates')
    env = jinja2.Environment(loader=loader)
    return env


def csv_generator_p2(records, fields, include_header=True, header=None,
                     dialect=csv.excel):

    def _row_string(row):
        writer.writerow(row)
        # Fetch UTF-8 output from the queue ...
        data = queue.getvalue()
        data = compat.to_unicode(data)
        # ... and reencode it into the target encoding
        data = encoder.encode(data)
        # empty queue
        queue.truncate(0)

        return data

    queue = compat.StringIO()
    writer = csv.writer(queue, dialect=dialect)
    encoder = codecs.getincrementalencoder("utf-8")()

    if include_header:
        yield _row_string(header or fields)

    for record in records:
        row = []
        for field in fields:
            value = record.get(field)
            if isinstance(value, compat.string_type):
                row.append(value.encode("utf-8"))
            elif value is not None:
                row.append(compat.text_type(value))
            else:
                row.append(None)

        yield _row_string(row)


def csv_generator_p3(records, fields, include_header=True, header=None,
                     dialect=csv.excel):

    def _row_string(row):
        writer.writerow(row)
        data = queue.getvalue()
        queue.truncate(0)

        return data

    queue = compat.StringIO()
    writer = csv.writer(queue, dialect=dialect)

    if include_header:
        yield _row_string(header or fields)

    for record in records:
        row = [record.get(field) for field in fields]

        yield _row_string(row)


if compat.py3k:
    csv_generator = csv_generator_p3
else:
    csv_generator = csv_generator_p2


class JSONLinesGenerator(object):
    def __init__(self, iterable, separator='\n'):
        """Creates a generator that yields one JSON record per record from
        `iterable` separated by a newline character.."""
        self.iterable = iterable
        self.separator = separator

        self.encoder = SlicerJSONEncoder(indent=None)

    def __iter__(self):
        for obj in self.iterable:
            string = self.encoder.encode(obj)
            yield u"{}{}".format(string, self.separator)


class SlicerJSONEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        """Creates a JSON encoder that will convert some data values and also allows
        iterables to be used in the object graph.

        :Attributes:
        * `iterator_limit` - limits number of objects to be fetched from
          iterator. Default: 1000.
        """

        super(SlicerJSONEncoder, self).__init__(*args, **kwargs)

        self.iterator_limit = 1000

    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()
        if hasattr(o, "to_dict") and callable(getattr(o, "to_dict")):
            return o.to_dict()
        else:
            array = None
            try:
                # If it is an iterator, then try to construct array and limit number of objects
                iterator = iter(o)
                count = self.iterator_limit
                array = []
                for i, obj in enumerate(iterator):
                    array.append(obj)
                    if i >= count:
                        break
            except TypeError as e:
                # not iterable
                pass

            if array is not None:
                return array
            else:
                return json.JSONEncoder.default(self, o)


class Formatter(object):
    """Empty class for the time being. Currently used only for finding all
    built-in subclasses"""
    def __call__(self, *args, **kwargs):
        return self.format(*args, **kwargs)

    def format(self, *args, **kwargs):
        raise NotImplementedError("Subclasses are expected to implement "
                                  "the format() method")


# Main pre-formatting
#

CrossTable = namedtuple("CrossTable", ["columns", "rows", "data"])

def make_cross_table(result, onrows=None, oncolumns=None, aggregates_on=None):
    """
    Creates a cross table from a drilldown (might be any list of records).
    `onrows` contains list of attribute names to be placed at rows and
    `oncolumns` contains list of attribute names to be placet at columns.
    `aggregates_on` specifies where the aggregates will be incuded:

    * ``cells`` (default) – aggregates will be included in the data matrix
      cell
    * ``rows`` – there will be one row per aggregate per "on row" dimension
      member
    * ``columns`` – there will be one column per aggregate per "on column"
      dimension member

    Returns a named tuble with attributes:

    * `columns` - labels of columns. The tuples correspond to values of
      attributes in `oncolumns`.
    * `rows` - labels of rows as list of tuples. The tuples correspond to
      values of attributes in `onrows`.
    * `data` - list of aggregate data per row. Each row is a list of
      aggregate tuples.

    """

    if not result.drilldown:
        # TODO: we should at least create one-row/one-column table
        raise ArgumentError("Can't create cross-table without drilldown.")

    aggregates = result.aggregates

    matrix = {}
    row_hdrs = []
    column_hdrs = []

    agg_refs = [agg.ref for agg in aggregates]

    if aggregates_on is None or aggregates_on == "cells":
        for record in result.cells:
            # Get table coordinates
            hrow = tuple(record[f] for f in onrows)
            hcol = tuple(record[f] for f in oncolumns)

            if hrow not in row_hdrs:
                row_hdrs.append(hrow)
            if hcol not in column_hdrs:
                column_hdrs.append(hcol)

            matrix[(hrow, hcol)] = tuple(record[a] for a in agg_refs)

    else:
        for record in result.cells:
            # Get table coordinates
            base_hrow = [record[f] for f in onrows]
            base_hcol = [record[f] for f in oncolumns]

            for agg in aggregates:

                if aggregates_on == "rows":
                    hrow = tuple(base_hrow + [agg.label or agg.name])
                    hcol = tuple(base_hcol)

                elif aggregates_on == "columns":
                    hrow = tuple(base_hrow)
                    hcol = tuple(base_hcol + [agg.label or agg.name])

                if hrow not in row_hdrs:
                    row_hdrs.append(hrow)

                if hcol not in column_hdrs:
                    column_hdrs.append(hcol)

                matrix[(hrow, hcol)] = record[agg.ref]

    data = []

    for hrow in row_hdrs:
        row = [matrix.get((hrow, hcol)) for hcol in column_hdrs]
        data.append(row)

    return CrossTable(column_hdrs, row_hdrs, data)


def coalesce_table_labels(attributes, onrows, oncolumns):
    """Returns a tuple 9`onrows`, `oncolumns`) containing `attributes`. If
    both are empty, all attributes will be put on rows. If one of the two is
    empty, the rest of attributes is put on that axis."""
    if not onrows or not oncolumns:
        onrows = onrows or []
        oncolumns = oncolumns or []

        if not onrows:
            onrows = [attr for attr in attributes if attr not in oncolumns]

        if not oncolumns:
            oncolumns = [attr for attr in attributes if attr not in onrows]

    return(onrows, oncolumns)


class CrossTableFormatter(Formatter):
    __options__ = [
                {
                    "name": "indent",
                    "type": "integer",
                    "label": "Output indent"
                },
            ]

    mime_type = "application/json"

    def __init__(self, indent=None):
        """Creates a cross-table formatter for JSON output.

        Arguments:

        * `indent` – output indentation

        If aggregates are put on rows or columns, then respective row or
        column is added per aggregate. The data contains single aggregate
        values.

        If aggregates are put in the table as cells, then the data contains
        tuples of aggregates in the order as specified in the `aggregates`
        argument of `format()` method.
        """

        self.indent = indent or 4
        self.encoder = SlicerJSONEncoder(indent=indent)

    def format(self, cube, result, onrows=None, oncolumns=None, aggregates=None,
               aggregates_on=None):

        onrows, oncolumns = coalesce_table_labels(result.attributes,
                                                  onrows,
                                                  oncolumns)
        table = make_cross_table(result,
                                 onrows=onrows,
                                 oncolumns=oncolumns,
                                 aggregates_on=aggregates_on)

        d = {
            "columns": table.columns,
            "rows": table.rows,
            "data": table.data
        }
        output = self.encoder.encode(d)

        return output


class HTMLCrossTableFormatter(CrossTableFormatter):
    __options__ = [
                {
                    "name": "table_style",
                    "description": "CSS style for the table"
                }
            ]
    mime_type = "text/html"

    def __init__(self, table_style=None):
        """Create a simple HTML table formatter. See `CrossTableFormatter` for
        information about arguments."""

        self.env = _jinja_env()
        self.template = self.env.get_template("cross_table.html")
        self.table_style = table_style

    def format(self, cube, result, onrows=None, oncolumns=None, aggregates=None,
                aggregates_on=None):

        onrows, oncolumns = coalesce_table_labels(result.attributes,
                                                  onrows,
                                                  oncolumns)
        table = make_cross_table(result,
                                 onrows=onrows,
                                 oncolumns=oncolumns,
                                 aggregates_on=aggregates_on)

        output = self.template.render(table=table,
                                      table_style=self.table_style)
        return output

class CSVFormatter(Formatter):
    def format(self, cube, result, onrows=None, oncolumns=None, aggregates=None,
               aggregates_on=None):

        if any([onrows, oncolumns]):
            raise ArgumentError("Column/row layout options are not supported")

        header = []
        for l in result.labels:
            # TODO: add a little bit of polish to this
            if l == SPLIT_DIMENSION_NAME:
                header.append('Matches Filters')
            else:
                header += [attr.label or attr.name
                           for attr in cube.get_attributes([l], aggregated=True)]

        fields = result.labels
        generator = csv_generator(result,
                                  fields,
                                  include_header=bool(header),
                                  header=header)
        # TODO: this is Py3 hack over Py2 hack
        rows = [compat.to_str(row) for row in generator]
        output = "".join(rows)
        return output




# -*- encoding: utf-8 -*-
"""Slicer – Cubes command-line tool

For more information run: slicer --help

To enable full user exception debugging set the ``CUBES_ERROR_DEBUG``
environment variable.
"""

from __future__ import absolute_import
from __future__ import print_function

import json
import os
import sys
import click
import cubes

from .. import compat

from ..datastructures import AttributeDict
from ..errors import InconsistencyError, ArgumentError, InternalError, UserError
from ..formatters import csv_generator, SlicerJSONEncoder, JSONLinesGenerator
from ..metadata import read_model_metadata, write_model_metadata_bundle
from ..workspace import Workspace
from ..errors import CubesError
from ..server import run_server
from ..server.base import read_slicer_config

from .. import ext

from ..cells import cuts_from_string, Cell
from ..browser import string_to_dimension_level


DEFAULT_CONFIG = "slicer.ini"


@click.group()
@click.pass_context
@click.option('--debug/--no-debug',
              envvar='CUBES_DEBUG',
              default=False,
              help="Enable/disable debugging output")
def cli(ctx, debug):
    ctx.obj = AttributeDict()
    ctx.obj.debug = debug


################################################################################
# Command: serve

@cli.command()
@click.argument('config', type=click.Path(exists=True), default=DEFAULT_CONFIG)
@click.option('--visualizer',
              help="Visualizer URL for /visualizer path")
@click.pass_context
def serve(ctx, config, visualizer):
    """Run Slicer HTTP server."""
    config = read_config(config)

    # FIXME: "visualizer" shouldn't be in "server" section
    if visualizer:
        config.set("server", "visualizer", visualizer)

    run_server(config, debug=ctx.obj.debug)

################################################################################
# Command: extension

@cli.command("ext-info")
@click.argument('extension_type', metavar='TYPE', required=False, default='all')
@click.argument('extension_name', metavar='NAME', required=False)
@click.pass_context
def extension_info(ctx, extension_type, extension_name):
    """Show info about Cubes extensions"""

    if extension_type == 'all':
        types = ext.EXTENSION_TYPES.items()
    else:
        label = ext.EXTENSION_TYPES[extension_type]
        types = [(extension_type, label)]

    if extension_name:
        # Print detailed extension information
        manager = getattr(ext, extension_type)
        extension = manager.get(extension_name)
        extension.load()

        desc = extension.description or "No description."
        click.echo("{e.name} - {e.label}\n\n"
                   "{desc}\n"
                   .format(e=extension, desc=desc))

        if extension.options:
            click.echo("Configuration options:\n")

            for option in extension.options.values():
                name = option.get("name")
                desc = option.get("description", option.get("label"))
                desc = " - {}".format(desc) if desc else ""
                type_ = option.get("type_", "string")

                click.echo("    {name} ({type_}){desc}"
                           .format(name=name, desc=desc,
                                   type_=type_))
        else:
            click.echo("No known options.")
    else:
        # List extensions
        click.echo("Available Cubes extensions:\n")
        for ext_type, _ in types:
            manager = getattr(ext, ext_type)
            manager.discover()
            names = manager.names()

            click.echo("{}:\n    {}\n".format(ext_type, ", ".join(names)))

    click.echo()

################################################################################
# Command: list

@cli.command()
@click.option('--verbose/--terse', 'verbose', default=False,
              help='Display also cube description')
@click.argument('config', required=False,
                default=DEFAULT_CONFIG, type=click.Path(exists=True))
@click.pass_context
def list(ctx, config, verbose):
    """List cubes"""
    ws = Workspace(config)

    for cube in ws.list_cubes():
        name = cube["name"]
        label = cube.get("label", name)
        desc = cube.get("description", "(no description)")
        if verbose:
            print("{} - {}\n    {}\n".format(name, label, desc))
        else:
            print("{} - {}".format(name, label))


################################################################################
# Command: valdate_model

@cli.group()
@click.pass_context
def model(ctx):
    """Model metadata tools."""
    pass

@model.command()
@click.option('--defaults', '-d', 'show_defaults', default=False,
              help='show defaults')
@click.option('--warnings/--no-warnings', 'show_warnings', default=True,
              help='enable/disable warnings')
@click.argument('model_path', metavar='MODEL')
def validate(show_defaults, show_warnings, model_path):
    """Validate model metadata"""

    click.echo("Reading model %s" % model_path)
    model = cubes.read_model_metadata(model_path)

    click.echo("Validating model...")
    result = cubes.providers.validate_model(model)

    error_count = 0
    warning_count = 0
    default_count = 0

    for error in result:
        if error.scope == "model":
            scope = "model"
        else:
            if error.object:
                scope = "%s '%s'" % (error.scope, error.object)
            else:
                scope = "unknown %s" % error.scope

        if error.property:
            scope += " property '%s'" % error.property

        show = True
        if error.severity == "error":
            error_count += 1
        elif error.severity == "warning":
            warning_count += 1
            show = show_warnings
        elif error.severity == "default":
            show = show_defaults
            default_count += 1

        if show:
            print("%s in %s: %s"
                  % (error.severity.upper(), scope, error.message))

    if error_count == 0:
        if warning_count == 0:
            if default_count == 0:
                verdict = "model can be used"
            else:
                verdict = "model can be used, " \
                          "make sure that the defaults reflect reality"
        else:
            verdict = "not recommended to use the model, " \
                      "some issues might emerge"
    else:
        verdict = "model can not be used"

    print("")
    print("Defaults used  %d" % default_count)
    print("Warning        %d" % warning_count)
    print("Errors         %d" % error_count)
    print("Summary        %s" % verdict)

    if error_count > 0:
        exit(1)

################################################################################
# Command: test

@cli.command()
@click.option('--aggregate', is_flag=True, default=False,
              help="Test aggregate of whole cube")
@click.option('--exclude-store', '-E', 'exclude_stores', multiple=True)
@click.option('--store', 'include_stores', multiple=True)
@click.argument('config', default=DEFAULT_CONFIG)
@click.argument('cube', nargs=-1)
def test(aggregate, exclude_stores, include_stores, config, cube):
    """Test every cube in the model"""
    workspace = cubes.Workspace(config)

    errors = []

    if cube:
        cube_list = cube
    else:
        cube_list = [c["name"] for c in workspace.list_cubes()]

    exclude = exclude_stores or []
    include = include_stores or []

    tested = 0

    for name in cube_list:
        cube = workspace.cube(name)

        click.echo("testing {}: ".format(name), nl=False)

        if cube.store_name in exclude \
                or (include and cube.store_name not in include):
            click.echo("pass")
            continue

        try:
            browser = workspace.browser(name)
        except Exception as e:
            errors.append((name, e))
            click.echo("BROWSER ERROR")
            continue

        tested += 1

        try:
            facts = browser.test(aggregate=aggregate)
        except NotImplementedError:
            click.echo("pass - no test")
        # FIXME XXX CubesError not defined
        except CubesError as e:
            errors.append((name, e))
            click.echo("ERROR")

    click.echo()
    click.echo("tested %d cubes" % tested)

    if errors:
        click.echo("%d ERRORS:" % len(errors))
        for (cube, e) in errors:
            if hasattr(e, "error_type"):
                etype = e.error_type
            else:
                etype = str(type(e))

            click.echo("%s: %s - %s" % (cube, etype, str(e)))
    else:
        click.echo("test passed")


@model.command()
@click.option('--format', 'model_format', type=click.Choice(["json", "bundle"]),
              default='json',
              help='output model format')
@click.option('--force', is_flag=True,
              default=False,
              help='replace existing model bundle')
@click.argument('model_path', metavar='MODEL')
@click.argument('target', required=False)
@click.pass_context
def convert(ctx, model_format, force, model_path, target):
    """Convert model between model formats."""

    metadata = read_model_metadata(model_path)
    if model_format == "json":
        if not target:
            print(json.dumps(metadata, indent=4))
        else:
            with open(target, "w") as f:
                json.dump(metadata, f, indent=4)
    elif model_format == "bundle":
        write_model_metadata_bundle(target, metadata, replace=force)

def read_config(cfg):
    """Read the configuration file."""
    return read_slicer_config(cfg)

################################################################################
# Group: sql

@cli.group()
@click.pass_context
@click.option('--store', nargs=1,
              help="Name of the store to use other than default. Must be SQL.")
@click.option('--config', nargs=1, default=DEFAULT_CONFIG,
              help="Name of slicer.ini configuration file")
def sql(ctx, store, config):
    """SQL store commands"""
    ctx.obj.workspace = cubes.Workspace(config)
    ctx.obj.store = ctx.obj.workspace.get_store(store)

################################################################################
# Command: sql denormalize

@sql.command()
@click.option('--force', is_flag=True, default=False,
              help='replace existing views')
@click.option('--materialize', '-m', is_flag=True, default=False,
              help='create materialized view (table)')
@click.option('--index/--no-index', default=True,
              help='create index for key attributes')
@click.option('--schema', '-s',
              help='target view schema (overrides default fact schema')
@click.argument('cube', required=False)
@click.argument('target', required=False)
@click.pass_context
def denormalize(ctx, force, materialize, index, schema, cube, target):
    """Create denormalized view(s) from cube(s)."""

    if not materialize and index:
        raise ArgumentError("Non-materialized views can't be indexed")

    # Shortcuts
    workspace = ctx.obj.workspace
    store = ctx.obj.store

    if cube:
        target = target or store.naming.denormalized_table_name(cube)
        cubes = [(cube, target)]
    else:
        names = workspace.cube_names()
        targets = [store.naming.denormalized_table_name(name)
                   for name in names]
        cubes = zip(names, targets)

    for cube_name, target in cubes:
        cube = workspace.cube(cube_name)
        store = workspace.get_store(cube.store_name or "default")

        print("denormalizing cube '%s' into '%s'" % (cube_name,
                                                     target))

        store.create_denormalized_view(cube, target,
                                       materialize=materialize,
                                       replace=force,
                                       create_index=index,
                                       keys_only=False,
                                       schema=schema)


# TODO: Nice to have it back
# @sql.command("ddl")
# @click.argument('cubes', required=False, nargs=-1)
# @click.pass_context
# def generate_ddl(ctx, cubes):
#     # Shortcuts
#     workspace = ctx.obj.workspace
#     store = ctx.obj.store
#
#     ddl = store.ddl_for_model(args.url, model, fact_prefix=args.fact_prefix,
#                                 dimension_prefix=args.dimension_prefix,
#                                 fact_suffix=args.fact_suffix,
#                                 dimension_suffix=args.dimension_suffix)
#
#     print(ddl)


################################################################################
# Command: sql aggregate

@sql.command("aggregate")
@click.option('--force', is_flag=True, default=False,
              help='replace existing views')
@click.option('--index/--no-index', default=True,
              help='create index for key attributes')
@click.option('--schema', '-s',
              help='target view schema (overrides default fact schema')
@click.option('--dimension', '-d', "dimensions", multiple=True,
              help='dimension to be used for aggregation')
@click.argument('cube', required=False)
@click.argument('target', required=False)
@click.pass_context
def sql_aggregate(ctx, force, index, schema, cube, target, dimensions):
    """Create pre-aggregated table from cube(s). If no cube is specified, then
    all cubes are aggregated. Target table can be specified only for one cube,
    for multiple cubes naming convention is used.
    """
    workspace = ctx.obj.workspace
    store = ctx.obj.store

    if cube:
        target = target or store.naming.aggregated_table_name(cube)
        cubes = [(cube, target)]
    else:
        names = workspace.cube_names()
        targets = [store.naming.aggregated_table_name(name)
                   for name in names]
        cubes = zip(names, targets)

    for cube_name, target in cubes:
        cube = workspace.cube(cube_name)
        store = workspace.get_store(cube.store_name or "default")

        print("denormalizing cube '%s' into '%s'" % (cube_name,
                                                     target))

        store.create_cube_aggregate(cube, target,
                                    replace=force,
                                    create_index=index,
                                    schema=schema,
                                    dimensions=dimensions)


################################################################################
# Command: aggregate

@cli.command()
@click.option('--config', type=click.Path(exists=True), required=False,
              default=DEFAULT_CONFIG)

@click.option('--aggregate', '-a', 'aggregates', multiple=True,
              help="List of aggregates to get")
@click.option('--cut', '-c', 'cuts', multiple=True,
              help="Cell cut")
@click.option('--split', 'split_str', multiple=False,
              help="Split cell")
@click.option('--drilldown', '-d', 'drilldown', multiple=True,
              help="Drilldown dimensions")

@click.option('--on-row', 'on_rows', multiple=True,
              help="Attribute to put on row (default is all)")
@click.option('--on-column', 'on_columns', multiple=True,
              help="Attribute to put on column (default is none)")

@click.option('--format', "-f", "formatter_name", default="cross_table",
              help="Output format")

@click.argument('cube_name', metavar='CUBE')
@click.pass_context
def aggregate(ctx, config, cube_name, aggregates, cuts, drilldown, formatter_name,
              split_str, on_rows, on_columns):
    """Aggregate a cube"""
    config = read_config(config)
    workspace = Workspace(config)
    browser = workspace.browser(cube_name)

    cell_cuts = []
    for cut_str in cuts:
        cell_cuts += cuts_from_string(browser.cube, cut_str)

    cell = Cell(browser.cube, cell_cuts)

    split_cuts = cuts_from_string(browser.cube, split_str)
    if split_cuts:
        split = Cell(browser.cube, split_cuts)
    else:
        split = None

    if not aggregates:
        aggregates = [agg.name for agg in browser.cube.aggregates]

    # TODO: paging and ordering
    result = browser.aggregate(cell,
                               aggregates=aggregates,
                               drilldown=drilldown,
                               split=split,
                               page=None,
                               page_size=None,
                               order=None)

    if formatter_name:
        formatter = ext.formatter(formatter_name)
        output = formatter.format(browser.cube,
                                  result,
                                  onrows=on_rows,
                                  oncolumns=on_columns,
                                  aggregates=aggregates,
                                  aggregates_on="columns")
    else:
        output = result.to_dict()

    click.echo(output)


################################################################################
# Command: members

@cli.command()
@click.option('--config', type=click.Path(exists=True), required=False,
              default=DEFAULT_CONFIG)

@click.option('--cut', '-c', 'cuts', multiple=True,
              help="Cell cut")

@click.option('--format', "-f", "output_format", default="json",
              type=click.Choice(["json", "csv", "json_lines" ]),
              help="Output format")

@click.argument('cube_name', metavar='CUBE')
@click.argument('dim_name', metavar='DIMENSION')
@click.pass_context
def members(ctx, config, cube_name, cuts, dim_name, output_format):
    """Aggregate a cube"""
    config = read_config(config)
    workspace = Workspace(config)
    browser = workspace.browser(cube_name)
    cube = browser.cube

    cell_cuts = []
    for cut_str in cuts:
        cell_cuts += cuts_from_string(browser.cube, cut_str)

    cell = Cell(browser.cube, cell_cuts)

    (dim_name, hier_name, level_name) = string_to_dimension_level(dim_name)
    dimension = cube.dimension(dim_name)
    hierarchy = dimension.hierarchy(hier_name)

    if level_name:
        depth = hierarchy.level_index(level_name) + 1
    else:
        depth = len(hierarchy)

    # TODO: pagination
    values = browser.members(cell,
                             dimension,
                             depth=depth,
                             hierarchy=hierarchy,
                             page=None,
                             page_size=None)

    attributes = []
    for level in hierarchy.levels_for_depth(depth):
        attributes += level.attributes

    fields = [attr.ref for attr in attributes]
    labels = [attr.label or attr.name for attr in attributes]

    if output_format == "json":
        encoder = SlicerJSONEncoder(indent=4)
        result = encoder.iterencode(values)
    elif output_format == "json_lines":
        result = JSONLinesGenerator(values)
    elif output_format == "csv":
        result = csv_generator(values,
                               fields,
                               include_header=True,
                               header=labels)

    out = click.get_text_stream('stdout')
    for row in result:
        out.write(row)


def main(*args, **kwargs):

    try:
        cli(*args, **kwargs)

    except InconsistencyError as e:
        # Internal Error - error caused by some edge case conditio, misbehaved
        # cubes or wrongly categorized error
        #
        # It is very unlikely that the user might fix this error by changing
        # his/her input.
        #
        if os.environ.get("CUBES_ERROR_DEBUG"):
            raise
        else:
            click.echo("\n" \
                  "Error: Internal error occured.\n"
                  "Reason: {}\n\n" \
                  "Please report the error and information about what you " \
                  "were doing to the Cubes development team.\n"
                  .format(e), err=True)
            sys.exit(1)

    except (InternalError, UserError) as e:
        # Error caused by the user – model or data related.
        #
        # User can fix the error by altering his/her input.
        #
        if os.environ.get("CUBES_ERROR_DEBUG"):
            raise
        else:
            click.echo("\nError: {}".format(e), err=True)
            sys.exit(1)




# -*- coding: utf-8 -*-
from contextlib import contextmanager
from collections import namedtuple
from threading import Thread

import datetime
import time
import csv
import io
import json

from .. import ext
from .. import compat
from ..logging import get_logger
from ..errors import *
from ..browser import Drilldown

__all__ = [
    "create_request_log_handler",
    "configured_request_log_handlers",

    "RequestLogger",
    "AsyncRequestLogger",
    "RequestLogHandler",
    "DefaultRequestLogHandler",
    "CSVFileRequestLogHandler",
    "QUERY_LOG_ITEMS"
]


REQUEST_LOG_ITEMS = [
    "timestamp",
    "method",
    "cube",
    "cell",
    "identity",
    "elapsed_time",
    "attributes",
    "split",
    "drilldown",
    "page",
    "page_size",
    "format",
    "headers"
]


def configured_request_log_handlers(config, prefix="query_log",
                                    default_logger=None):
    """Returns configured query loggers as defined in the `config`."""

    handlers = []

    for section in config.sections():
        if section.startswith(prefix):
            options = dict(config.items(section))
            type_ = options.pop("type")
            if type_ == "default":
                logger = default_logger or get_logger()
                handler = ext.request_log_handler("default", logger)
            else:
                handler = ext.request_log_handler(type_, **options)

            handlers.append(handler)

    return handlers


class RequestLogger(object):
    def __init__(self, handlers=None):
        if handlers:
            self.handlers = list(handlers)
        else:
            self.handlers = []

        self.logger = get_logger()

    @contextmanager
    def log_time(self, method, browser, cell, identity=None, **other):
        start = time.time()
        yield
        elapsed = time.time() - start
        self.log(method, browser, cell, identity, elapsed, **other)

    def log(self, method, browser, cell, identity=None, elapsed=None, **other):

        record = {
            "timestamp": datetime.datetime.now(),
            "method": method,
            "cube": browser.cube,
            "identity": identity,
            "elapsed_time": elapsed or 0,
            "cell": cell
        }
        record.update(other)

        record = self._stringify_record(record)

        for handler in self.handlers:
            try:
                handler.write_record(browser.cube, cell, record)
            except Exception as e:
                self.logger.error("Server log handler error (%s): %s"
                                  % (type(handler).__name__, str(e)))


    def _stringify_record(self, record):
        """Return a log rectord with object attributes converted to unicode strings"""
        record = dict(record)

        record["cube"] = compat.text_type(record["cube"])

        cell = record.get("cell")
        record["cell"] = compat.text_type(cell) if cell is not None else None

        split = record.get("split")
        record["split"] = compat.text_type(split) if split is not None else None

        return record


class AsyncRequestLogger(RequestLogger):
    def __init__(self, handlers=None):
        super(AsyncRequestLogger, self).__init__(handlers)
        self.queue = compat.Queue()
        self.thread = Thread(target=self.log_consumer,
                              name="slicer_logging")
        self.thread.daemon = True
        self.thread.start()

    def log(self, *args, **kwargs):
        self.queue.put( (args, kwargs) )

    def log_consumer(self):
        while True:
            (args, kwargs) = self.queue.get()
            super(AsyncRequestLogger, self).log(*args, **kwargs)

class RequestLogHandler(object):
    def write_record(self, record):
        pass


class DefaultRequestLogHandler(RequestLogHandler):
    def __init__(self, logger=None, **options):
        self.logger = logger

    def write_record(self, cube, cell, record, **options):
        if cell:
            cell_str = "'%s'" % str(cell)
        else:
            cell_str = "none"

        if record.get("identity"):
            identity_str = "'%s'" % str(record["identity"])
        else:
            identity_str = "none"

        self.logger.info("method:%s cube:%s cell:%s identity:%s time:%s"
                         % (record["method"], record["cube"], cell_str,
                            identity_str, record["elapsed_time"]))


class CSVFileRequestLogHandler(RequestLogHandler):
    def __init__(self, path=None, **options):
        self.path = path

    def write_record(self, cube, cell, record):
        out = []

        for key in REQUEST_LOG_ITEMS:
            item = record.get(key)
            if item is not None:
                item = compat.text_type(item)
            out.append(item)

        with io.open(self.path, 'ab') as f:
            writer = csv.writer(f)
            writer.writerow(out)

class JSONRequestLogHandler(RequestLogHandler):
    def __init__(self, path=None, **options):
        """Creates a JSON logger which logs requests in a JSON lines. It
        includes two lists: `cell_dimensions` and `drilldown_dimensions`."""
        self.path = path

    def write_record(self, cube, cell, record):
        out = []

        drilldown = record.get("drilldown")

        if drilldown is not None:
            if cell:
                drilldown = Drilldown(drilldown, cell)
                record["drilldown"] = str(drilldown)
            else:
                drilldown = []
                record["drilldown"] = None

        record["timestamp"] = record["timestamp"].isoformat()
        # Collect dimension uses
        uses = []

        cuts = cell.cuts if cell else []

        for cut in cuts:
            dim = cube.dimension(cut.dimension)
            depth = cut.level_depth()
            if depth:
                level = dim.hierarchy(cut.hierarchy)[depth-1]
                level_name = str(level)
            else:
                level_name = None

            use = {
                "dimension": str(dim),
                "hierarchy": str(cut.hierarchy),
                "level": str(level_name),
                "value": str(cut)
            }
            uses.append(use)

        record["cell_dimensions"] = uses

        uses = []

        for item in drilldown or []:
            (dim, hier, levels) = item[0:3]
            if levels:
                level = str(levels[-1])
            else:
                level = None

            use = {
                "dimension": str(dim),
                "hierarchy": str(hier),
                "level": str(level),
                "value": None
            }
            uses.append(use)

        record["drilldown_dimensions"] = uses
        line = json.dumps(record)

        with io.open(self.path, 'ab') as f:
            json.dump(record, f)
            f.write("\n")




"""OLAP Cubes"""

__version__ = "1.1"

from .common import *
from .browser import *
from .cells import *
from .model import *
from .metadata import *
from .workspace import *
from .errors import *
from .formatters import *
from .computation import *
from .mapper import *
from .providers import *
from .calendar import *
from .auth import *
from .logging import *
from .namespace import *





# -*- coding: utf-8 -*-
"""Logical to Physical Mappers"""

import collections

from .logging import get_logger

__all__ = (
    "Mapper",
)

class Mapper(object):
    """Mapper is core class for translating logical model to physical database
    schema.
    """
    # WARNING: do not put any SQL/engine/connection related stuff into this
    # class yet. It might be moved to the cubes as one of top-level modules
    # and subclassed here.

    def __init__(self, cube, locale=None, **naming):
        """Abstract class for mappers which maps logical references to
        physical references (tables and columns).

        Attributes:

        * `cube` - mapped cube
        * `fact_name` – fact name, if not specified then `cube.name` is used
        * `schema` – default database schema

        """

        super(Mapper, self).__init__()

        if cube is None:
            raise Exception("Cube for mapper should not be None.")

        self.logger = get_logger()

        self.cube = cube

        self.mappings = self.cube.mappings
        self.locale = locale

        # TODO: remove this (should be in SQL only)

        self._collect_attributes()

    def _collect_attributes(self):
        """Collect all cube attributes and create a dictionary where keys are
        logical references and values are `cubes.model.Attribute` objects.
        This method should be used after each cube or mappings change.
        """

        self.attributes = collections.OrderedDict()

        for attr in self.cube.all_fact_attributes:
            self.attributes[self.logical(attr)] = attr

    def set_locale(self, locale):
        """Change the mapper's locale"""
        self.locale = locale
        self._collect_attributes()

    # TODO: depreciate in favor of Cube.all_attributes
    def all_attributes(self, expand_locales=False):
        """Return a list of all attributes of a cube. If `expand_locales` is
        ``True``, then localized logical reference is returned for each
        attribute's locale."""
        return self.attributes.values()

    # TODO: depreciate in favor of Cube.attribute
    def attribute(self, name):
        """Returns an attribute with logical reference `name`. """
        # TODO: If attribute is not found, returns `None` (yes or no?)

        return self.attributes[name]

    # TODO: is this necessary after removing of 'simplify'? Reconsider
    # requirement for existence of this one.
    def logical(self, attribute, locale=None):
        """Returns logical reference as string for `attribute` in `dimension`.
        If `dimension` is ``Null`` then fact table is assumed. The logical
        reference might have following forms:

        * ``dimension.attribute`` - dimension attribute
        * ``attribute`` - fact measure or detail

        If `locale` is specified, then locale is added to the reference. This
        is used by backends and other mappers, it has no real use in end-user
        browsing.
        """

        reference = attribute.localized_ref(locale)

        return reference

    def split_logical(self, reference):
        """Returns tuple (`dimension`, `attribute`) from `logical_reference` string. Syntax
        of the string is: ``dimensions.attribute``."""

        split = reference.split(".")

        if len(split) > 1:
            dim_name = split[0]
            attr_name = ".".join(split[1:])
            return (dim_name, attr_name)
        else:
            return (None, reference)

    def physical(self, attribute, locale=None):
        """Returns physical reference for attribute. Returned value is backend
        specific. Default implementation returns a value from the mapping
        dictionary.

        This method should be implemented by `Mapper` subclasses.
        """

        return self.mappings.get(self.logical(attribute, locale))




# -*- coding: utf-8 -*-
from collections import OrderedDict
from textwrap import dedent
from pkg_resources import iter_entry_points

from .common import decamelize, coalesce_options
from .errors import ArgumentError, InternalError, BackendError


__all__ = [
    "EXTENSION_TYPES",
    "ExtensionFinder",
]

# Known extension types.
# Keys:
#     base: extension base class name
#     suffix: extension class suffix to be removed for default name (same as
#         base class nameif not specified)
#     modules: a dictionary of extension names and module name to be loaded
#         laily


EXTENSION_TYPES = {
    "browser": "Aggregation browser",
    "store": "Data store",
    "model_provider": "Model provider",
    "formatter": "Formatter",
    "authorizer": "Authorizer",
    "authenticator": "Authenticator",
    "request_log_handler": "Request log handler",
}

# Information about built-in extensions. Supposedly faster loading (?).
#
_BUILTIN_EXTENSIONS = {
    "authenticators": {
        "admin_admin": "cubes.server.auth:AdminAdminAuthenticator",
        "pass_parameter": "cubes.server.auth:PassParameterAuthenticator",
        "http_basic_proxy": "cubes.server.auth:HTTPBasicProxyAuthenticator",
    },
    "authorizers": {
        "simple": "cubes.auth:SimpleAuthorizer",
    },
    "browsers": {
        "sql":"cubes.sql.browser:SQLBrowser",
        "slicer":"cubes.server.browser:SlicerBrowser",
    },
    "formatters": {
        "cross_table": "cubes.formatters:CrossTableFormatter",
        "csv": "cubes.formatters:CSVFormatter",
        "html_cross_table": "cubes.formatters:HTMLCrossTableFormatter",
    },
    "providers": {
        "default":"cubes.providers:StaticModelProvider",
        "slicer":"cubes.server.store:SlicerModelProvider",
    },
    "request_log_handlers": {
        "default":"cubes.server.logging:DefaultRequestLogHandler",
        "csv":"cubes.server.logging:CSVFileRequestLogHandler",
        "json":"cubes.server.logging:JSONRequestLogHandler",
        "sql":"cubes.sql.logging:SQLRequestLogger",
    },
    "stores": {
        "sql":"cubes.sql.store:SQLStore",
        "slicer":"cubes.server.store:SlicerStore",
    },
}

_DEFAULT_OPTIONS = {
}

class _Extension(object):
    """
    Cubes Extension wrapper.

    `options` – List of extension options.  The options is a list of
    dictionaries with keys:

    * `name` – option name
    * `type` – option data type (default is ``string``)
    * `description` – description (optional)
    * `label` – human readable label (optional)
    * `values` – valid values for the option.
    """
    def __init__(self, type_, entry=None, factory=None, name=None):
        if factory is not None and entry is not None:
            raise ArgumentError("Can't set both extension factory and entry "
                                "(in extension '{}')".format(name))

        elif factory is None and entry is None:
            raise ArgumentError("Neither extension factory nor entry provided "
                                "(in extension '{}')".format(name))

        self.type_ = type_
        self.entry = entry
        self.name = name or entry.name

        # After loading...
        self.options = []
        self.option_types = {}
        self._factory = None

        if factory is not None:
            self.factory = factory

    @property
    def factory(self):
        if self._factory is not None:
            return self._factory
        elif self.entry:
            # This must not fail or result in None
            self.factory = self.entry.load()
            return self._factory
        else:
            raise InternalError("No factory or entry set for extension '{}'"
                                .format(self.name))

    @factory.setter
    def factory(self, factory):
        if factory is None:
            raise InternalError("Can't set extension factory to None")

        self._factory = factory
        defaults = _DEFAULT_OPTIONS.get(self.type_, [])

        if hasattr(self._factory, "__options__"):
            options = self._factory.__options__ or []
        else:
            options = []

        self.options = OrderedDict()
        for option in defaults + options:
            name = option["name"]
            self.options[name] = option
            self.option_types[name] = option.get("type", "string")

        self.option_types = self.option_types or {}

    @property
    def is_builtin(self):
        return self.entry is None

    @property
    def label(self):
        if hasattr(self.factory, "__label__"):
            return self.factory.__label__
        else:
            return decamelize(self.factory.__name__)

    @property
    def description(self):
        if hasattr(self.factory, "__description__"):
            desc = self.factory.__description__ or ""
            return dedent(desc)
        else:
            return ""

    def create(self, *args, **kwargs):
        """Creates an extension. First argument should be extension's name."""
        factory = self.factory

        kwargs = coalesce_options(dict(kwargs),
                                  self.option_types)

        return factory(*args, **kwargs)


class ExtensionFinder(object):
    def __init__(self, type_):
        self.type_ = type_
        self.group = "cubes.{}".format(type_)
        self.extensions = {}

        self.builtins = _BUILTIN_EXTENSIONS.get(self.type_, {})

    def discover(self, name=None):
        """Find all entry points."""
        for obj in iter_entry_points(group=self.group, name=name):
            ext = _Extension(self.type_, obj)
            self.extensions[ext.name] = ext

    def builtin(self, name):
        try:
            ext_mod = self.builtins[name]
        except KeyError:
            return None

        (modname, attr) = ext_mod.split(":")
        module = _load_module(modname)
        factory = getattr(module, attr)
        ext = _Extension(self.type_, name=name, factory=factory)
        self.extensions[name] = ext

        return ext

    def names(self):
        """Return list of extension names."""
        if not self.extensions:
            self.discover()

        names = list(self.builtins.keys())
        names += self.extensions.keys()

        return sorted(names)

    def get(self, name):
        """Return extenson object by name. Load if necessary."""
        ext = self.extensions.get(name)

        if not ext:
            ext = self.builtin(name)

        if not ext:
            self.discover()

            try:
                ext = self.extensions[name]
            except KeyError:
                raise InternalError("Unknown '{}' extension '{}'"
                                    .format(self.type_, name))
        return ext

    def __call__(self, _ext_name, *args, **kwargs):
        return self.create(_ext_name, *args, **kwargs)

    def factory(self, name):
        """Return extension factory."""
        ext = self.get(name)

        if not ext.factory:
            raise BackendError("Unable to get factory for extension '{}'"
                               .format(name))

        return ext.factory

    def create(self, _ext_name, *args, **kwargs):
        """Create an instance of extension `_ext_name` with given arguments.
        The keyword arguments are converted to their appropriate types
        according to extensions `__options__` list. This allows options to be
        specified as strings in a configuration files or configuration
        variables."""
        ext = self.get(_ext_name)
        return ext.create(*args, **kwargs)

    def register(self, _ext_name, factory):
        ext = _Extension(self.type_, name=_ext_name)
        ext.factory = factory
        self.extensions["name"] = ext

        return ext


def _load_module(modulepath):
    """Load module `modulepath` and return the last module object in the
    module path."""

    mod = __import__(modulepath)
    path = []
    for token in modulepath.split(".")[1:]:
        path.append(token)
        mod = getattr(mod, token)
    return mod


authenticator = ExtensionFinder("authenticators")
authorizer = ExtensionFinder("authorizers")
browser = ExtensionFinder("browsers")
formatter = ExtensionFinder("formatters")
model_provider = ExtensionFinder("providers")
request_log_handler = ExtensionFinder("request_log_handlers")
store = ExtensionFinder("stores")
